// $Id$
//
// Copyright (C) 2010

/** \file
    \brief InternalNetworkAccessManager non-inline non-template implementation */

#include "Publisher.hh"
//#include "Publisher.ih"

// Custom includes
#include <iostream>
#include <QNetworkProxy>
#include <QVariant>
#include "Publisher.hh"
#include <boost/python.hpp>

//#include "Publisher.mpp"
#define prefix_
///////////////////////////////cc.p////////////////////////////////////////

///////////////////////////////////////////////////////////////////////////
// pykit::detail::InternalServerReply

namespace pykit {
namespace detail {

    class InternalServerReply
        : public QNetworkReply
    {
        Q_OBJECT;
    public:
        InternalServerReply(QNetworkAccessManager::Operation operation,
                            QNetworkRequest const & networkRequest,
                            QIODevice * postData, Publisher * publisher);

        // Forwarded to buffer
        virtual bool atEnd() const;
        virtual qint64 bytesAvailable() const;
        virtual qint64 bytesToWrite() const;
        virtual bool canReadLine() const;
        virtual bool isSequential() const;
        virtual qint64 pos() const;
        virtual bool reset();
        virtual bool seek(qint64 pos);
        virtual qint64 size() const;
        virtual bool waitForBytesWritten(int msecs);
        virtual bool waitForReadyRead(int msecs);

        virtual void abort();

        using QNetworkReply::setHeader;
        using QNetworkReply::setRawHeader;

        void clearResponse();
        QByteArray postData() const;
        QString postContentType() const;

    protected:
        virtual qint64 readData(char * data, qint64 maxSize);
        virtual qint64 writeData(char const * data, qint64 maxSize);

    private slots:
        void bufferPostData();
        void postDataReadable();
        void postDataComplete();
        void processRequest();
        void complete();

    private:
        bool readPostData();

        Publisher * publisher_;
        QIODevice * postDevice_;
        QBuffer buffer_;
        QBuffer postData_;
    };

}}

prefix_ pykit::detail::InternalServerReply::
InternalServerReply(QNetworkAccessManager::Operation operation,
                    QNetworkRequest const & networkRequest, QIODevice * postData,
                    Publisher * publisher)
    : publisher_ (publisher), postDevice_ (postData)
{
    setRequest(networkRequest);
    setOperation(operation);
    setUrl(networkRequest.url());
    open(ReadWrite | Unbuffered);
    buffer_.open(ReadWrite | Unbuffered);

    if (postDevice_)
        QMetaObject::invokeMethod(this, "bufferPostData", Qt::QueuedConnection);
    else
        QMetaObject::invokeMethod(this, "processRequest", Qt::QueuedConnection);
}

prefix_ void pykit::detail::InternalServerReply::bufferPostData()
{
    QVariant cl (request().header(QNetworkRequest::ContentLengthHeader));
    connect(postDevice_, SIGNAL(readyRead()), this, SLOT(postDataReadable()));
    connect(postDevice_, SIGNAL(readChannelFinished()), this, SLOT(postDataComplete()));
    postData_.open(WriteOnly);
    postDataReadable();
}

prefix_ bool pykit::detail::InternalServerReply::readPostData()
{
    while (true) {
        qint64 n (postDevice_->bytesAvailable());
        if (n <= 0)
            n = 2048;
        QByteArray data;
        data.resize(n);
        n  = postDevice_->read(data.data(), n);
        if (n < 0)
            return true;
        else {
            data.resize(n);
            postData_.write(data);
        }
        if (n == 0)
            return false;
    }
}

prefix_ void pykit::detail::InternalServerReply::postDataReadable()
{
    if (readPostData()) {
        disconnect(postDevice_, SIGNAL(readyRead()), this, SLOT(postDataReadable()));
        disconnect(postDevice_, SIGNAL(readChannelFinished()), this, SLOT(postDataComplete()));
        processRequest();
    }
}

prefix_ void pykit::detail::InternalServerReply::postDataComplete()
{
    readPostData();
    disconnect(postDevice_, SIGNAL(readyRead()), this, SLOT(postDataReadable()));
    disconnect(postDevice_, SIGNAL(readChannelFinished()), this, SLOT(postDataComplete()));
    processRequest();
}

prefix_ void pykit::detail::InternalServerReply::processRequest()
{
    Request request (*this);
    publisher_->publish(request);

    seek(0);
    setHeader(QNetworkRequest::ContentLengthHeader, QVariant(size()));

    QMetaObject::invokeMethod(this, "complete", Qt::QueuedConnection);
}

prefix_ void pykit::detail::InternalServerReply::complete()
{
    emit readyRead();
    emit readChannelFinished();
    emit finished();
}

prefix_ bool pykit::detail::InternalServerReply::atEnd()
    const
{ return buffer_.atEnd(); }

prefix_ qint64 pykit::detail::InternalServerReply::bytesAvailable()
    const
{ return buffer_.bytesAvailable(); }

prefix_ qint64 pykit::detail::InternalServerReply::bytesToWrite()
    const
{ return buffer_.bytesToWrite(); }

prefix_ bool pykit::detail::InternalServerReply::canReadLine()
    const
{ return buffer_.canReadLine(); }

prefix_ bool pykit::detail::InternalServerReply::isSequential()
    const
{ return buffer_.isSequential(); }

prefix_ qint64 pykit::detail::InternalServerReply::pos()
    const
{ return buffer_.pos(); }

prefix_ bool pykit::detail::InternalServerReply::reset()
{ return buffer_.reset(); }

prefix_ bool pykit::detail::InternalServerReply::seek(qint64 pos)
{ return buffer_.seek(pos); }

prefix_ qint64 pykit::detail::InternalServerReply::size()
    const
{ return buffer_.size(); }

prefix_ bool pykit::detail::InternalServerReply::waitForBytesWritten(int msecs)
{ return buffer_.waitForBytesWritten(msecs); }

prefix_ bool pykit::detail::InternalServerReply::waitForReadyRead(int msecs)
{ return buffer_.waitForReadyRead(msecs); }

prefix_ qint64 pykit::detail::InternalServerReply::readData(char * data, qint64 maxSize)
{ return buffer_.read(data,maxSize); }

prefix_ qint64 pykit::detail::InternalServerReply::writeData(char const * data, qint64 maxSize)
{ return buffer_.write(data,maxSize); }

prefix_ void pykit::detail::InternalServerReply::abort()
{}

prefix_ void pykit::detail::InternalServerReply::clearResponse()
{
    buffer_.buffer().clear();
}

prefix_ QByteArray pykit::detail::InternalServerReply::postData()
    const
{
    return postData_.buffer();
}

prefix_ QString pykit::detail::InternalServerReply::postContentType()
    const
{
    return request().header(QNetworkRequest::ContentTypeHeader).toString();
}

///////////////////////////////////////////////////////////////////////////
// pykit::Request

prefix_ void pykit::Request::write(std::string const & data)
{
    reply_.write(QByteArray(data.data(),data.size()));
}

prefix_ void pykit::Request::reset()
{
    reply_.clearResponse();
}

prefix_ void pykit::Request::setContentType(std::string const & contentType)
{
    reply_.setHeader(QNetworkRequest::ContentTypeHeader, QVariant(contentType.c_str()));
}

prefix_ void pykit::Request::setLocation(std::string const & location)
{
    reply_.setHeader(QNetworkRequest::LocationHeader, QVariant(location.c_str()));
}

prefix_ void pykit::Request::setHeader(std::string const & header, std::string const & value)
{
    reply_.setRawHeader(header.c_str(), value.c_str());
}

prefix_ QUrl pykit::Request::url()
    const
{
    return reply_.url();
}

prefix_ QByteArray pykit::Request::postData()
    const
{
    return reply_.postData();
}

prefix_ QByteArray pykit::Request::operation()
    const
{
    static char const * const operationNames[]
        = { "", "HEAD", "GET", "PUT", "POST", "DELETE", "" };
    return operationNames[reply_.operation()];
}

prefix_ QString pykit::Request::postContentType()
    const
{
    return reply_.postContentType();
}

prefix_ pykit::Request::Request(detail::InternalServerReply & reply)
    : reply_ (reply)
{}

///////////////////////////////////////////////////////////////////////////
// pykit::InternalNetworkAccessManager

prefix_ pykit::InternalNetworkAccessManager::
InternalNetworkAccessManager(QNetworkAccessManager * manager, QObject * parent,
                             Publisher * publisher)
    : QNetworkAccessManager(parent), publisher_ (publisher)
{
    setCache(manager->cache());
    setCookieJar(manager->cookieJar());
    setProxy(manager->proxy());
    setProxyFactory(manager->proxyFactory());
}

prefix_ QNetworkReply *
pykit::InternalNetworkAccessManager::createRequest(Operation operation,
                                                        QNetworkRequest const & request,
                                                        QIODevice * device)
{
    if (request.url().host() != "pykit")
        return QNetworkAccessManager::createRequest(operation, request, device);
    return new detail::InternalServerReply(operation, request, device, publisher_);
}

#include "Publisher.moc"

///////////////////////////////cc.e////////////////////////////////////////
#undef prefix_
//#include "Publisher.mpp"


// Local Variables:
// mode: c++
// fill-column: 100
// comment-column: 40
// c-file-style: "j32"
// indent-tabs-mode: nil
// ispell-local-dictionary: "american"
// compile-command: "scons -U"
// End:
